@@ -1,5 +1,5 @@
-use crate::frame_info::GameInput;
-use crate::{Frame, NULL_FRAME};
+use crate::frame_info::PlayerInput;
+use crate::{Config, Frame, InputStatus, NULL_FRAME};
 use std::cmp;
 
 /// The length of the input queue. This describes the number of inputs GGRS can hold at the same time per player.
@@ -7,7 +7,10 @@ const INPUT_QUEUE_LENGTH: usize = 128;
 
 /// `InputQueue` handles inputs for a single player and saves them in a circular array. Valid Inputs are between `head` and `tail`.
 #[derive(Debug, Clone)]
-pub(crate) struct InputQueue {
+pub(crate) struct InputQueue<T>
+where
+    T: Config,
+{
     /// The head of the queue. The newest `GameInput` is saved here      
     head: usize,
     /// The tail of the queue. The oldest `GameInput` still valid is saved here.
@@ -25,16 +28,16 @@ pub(crate) struct InputQueue {
     last_requested_frame: Frame,
 
     /// The delay in frames by which inputs are sent back to the user. This can be set during initialization.
-    frame_delay: u32,
+    frame_delay: usize,
 
     /// Our cyclic input queue
-    inputs: Vec<GameInput>,
+    inputs: Vec<PlayerInput<T::Input>>,
     /// A pre-allocated prediction we are going to use to return predictions from.
-    prediction: GameInput,
+    prediction: PlayerInput<T::Input>,
 }
 
-impl InputQueue {
-    pub(crate) fn new(input_size: usize) -> Self {
+impl<T: Config> InputQueue<T> {
+    pub(crate) fn new() -> Self {
         Self {
             head: 0,
             tail: 0,
@@ -44,16 +47,16 @@ impl InputQueue {
             last_added_frame: NULL_FRAME,
             first_incorrect_frame: NULL_FRAME,
             last_requested_frame: NULL_FRAME,
-            prediction: GameInput::blank_input(input_size),
-            inputs: vec![GameInput::blank_input(input_size); INPUT_QUEUE_LENGTH],
+            prediction: PlayerInput::blank_input(NULL_FRAME),
+            inputs: vec![PlayerInput::blank_input(NULL_FRAME); INPUT_QUEUE_LENGTH],
         }
     }
 
-    pub(crate) const fn first_incorrect_frame(&self) -> Frame {
+    pub(crate) fn first_incorrect_frame(&self) -> Frame {
         self.first_incorrect_frame
     }
 
-    pub(crate) fn set_frame_delay(&mut self, delay: u32) {
+    pub(crate) fn set_frame_delay(&mut self, delay: usize) {
         self.frame_delay = delay;
     }
 
@@ -65,11 +68,11 @@ impl InputQueue {
 
     /// Returns a `GameInput`, but only if the input for the requested frame is confirmed.
     /// In contrast to `input()`, this will not return a prediction if there is no confirmed input for the frame, but panic instead.
-    pub(crate) fn confirmed_input(&self, requested_frame: Frame) -> &GameInput {
+    pub(crate) fn confirmed_input(&self, requested_frame: Frame) -> PlayerInput<T::Input> {
         let offset = requested_frame as usize % INPUT_QUEUE_LENGTH;
 
         if self.inputs[offset].frame == requested_frame {
-            return &self.inputs[offset];
+            return self.inputs[offset];
         }
 
         // the requested confirmed input should not be before a prediction. We should not have asked for a known incorrect frame.
@@ -98,7 +101,7 @@ impl InputQueue {
     }
 
     /// Returns the game input of a single player for a given frame, if that input does not exist, we return a prediction instead.
-    pub(crate) fn input(&mut self, requested_frame: Frame) -> GameInput {
+    pub(crate) fn input(&mut self, requested_frame: Frame) -> (T::Input, InputStatus) {
         // No one should ever try to grab any input when we have a prediction error.
         // Doing so means that we're just going further down the wrong path. Assert this to verify that it's true.
         assert!(self.first_incorrect_frame == NULL_FRAME);
@@ -117,20 +120,20 @@ impl InputQueue {
             if offset < self.length {
                 offset = (offset + self.tail) % INPUT_QUEUE_LENGTH;
                 assert!(self.inputs[offset].frame == requested_frame);
-                return self.inputs[offset].clone();
+                return (self.inputs[offset].input, InputStatus::Confirmed);
             }
 
             // The requested frame isn't in the queue. This means we need to return a prediction frame. Predict that the user will do the same thing they did last time.
             if requested_frame == 0 || self.last_added_frame == NULL_FRAME {
                 // basing new prediction frame from nothing, since we are on frame 0 or we have no frames yet
-                self.prediction.erase_bits();
+                self.prediction = PlayerInput::blank_input(self.prediction.frame);
             } else {
                 // basing new prediction frame from previously added frame
                 let previous_position = match self.head {
                     0 => INPUT_QUEUE_LENGTH - 1,
                     _ => self.head - 1,
                 };
-                self.prediction = self.inputs[previous_position].clone();
+                self.prediction = self.inputs[previous_position];
             }
             // update the prediction's frame
             self.prediction.frame += 1;
@@ -138,13 +141,13 @@ impl InputQueue {
 
         // We must be predicting, so we return the prediction frame contents. We are adjusting the prediction to have the requested frame.
         assert!(self.prediction.frame != NULL_FRAME);
-        let mut prediction_to_return = self.prediction.clone(); // GameInput has copy semantics
+        let mut prediction_to_return = self.prediction; // GameInput has copy semantics
         prediction_to_return.frame = requested_frame;
-        prediction_to_return
+        (prediction_to_return.input, InputStatus::Predicted)
     }
 
     /// Adds an input frame to the queue. Will consider the set frame delay.
-    pub(crate) fn add_input(&mut self, input: GameInput) -> Frame {
+    pub(crate) fn add_input(&mut self, input: PlayerInput<T::Input>) -> Frame {
         // Verify that inputs are passed in sequentially by the user, regardless of frame delay.
         assert!(
             self.last_added_frame == NULL_FRAME
@@ -162,18 +165,17 @@ impl InputQueue {
 
     /// Adds an input frame to the queue at the given frame number. If there are predicted inputs, we will check those and mark them as incorrect, if necessary.
     /// Returns the frame number
-    fn add_input_by_frame(&mut self, input: GameInput, frame_number: Frame) {
+    fn add_input_by_frame(&mut self, input: PlayerInput<T::Input>, frame_number: Frame) {
         let previous_position = match self.head {
             0 => INPUT_QUEUE_LENGTH - 1,
             _ => self.head - 1,
         };
 
-        assert!(input.size == self.prediction.size);
         assert!(self.last_added_frame == NULL_FRAME || frame_number == self.last_added_frame + 1);
         assert!(frame_number == 0 || self.inputs[previous_position].frame == frame_number - 1);
 
         // Add the frame to the back of the queue
-        self.inputs[self.head] = input.clone();
+        self.inputs[self.head] = input;
         self.inputs[self.head].frame = frame_number;
         self.head = (self.head + 1) % INPUT_QUEUE_LENGTH;
         self.length += 1;
@@ -224,7 +226,7 @@ impl InputQueue {
         // This can occur when the frame delay has been increased since the last time we shoved a frame into the system.
         // We need to replicate the last frame in the queue several times in order to fill the space left.
         while expected_frame < input_frame {
-            let input_to_replicate = self.inputs[previous_position].clone();
+            let input_to_replicate = self.inputs[previous_position];
             self.add_input_by_frame(input_to_replicate, expected_frame);
             expected_frame += 1;
         }
@@ -245,35 +247,50 @@ impl InputQueue {
 #[cfg(test)]
 mod input_queue_tests {
 
+    use std::net::SocketAddr;
+
+    use bytemuck::{Pod, Zeroable};
+
     use super::*;
 
+    #[repr(C)]
+    #[derive(Copy, Clone, PartialEq, Pod, Zeroable)]
+    struct TestInput {
+        inp: u8,
+    }
+
+    struct TestConfig;
+
+    impl Config for TestConfig {
+        type Input = TestInput;
+        type State = Vec<u8>;
+        type Address = SocketAddr;
+    }
+
     #[test]
     #[should_panic]
     fn test_add_input_wrong_frame() {
-        let input_size = std::mem::size_of::<u32>();
-        let mut queue = InputQueue::new(input_size);
-        let input = GameInput::new(0, input_size, vec![0]);
+        let mut queue = InputQueue::<TestConfig>::new();
+        let input = PlayerInput::new(0, TestInput { inp: 0 });
         queue.add_input(input); // fine
-        let input_wrong_frame = GameInput::new(3, input_size, vec![0]);
+        let input_wrong_frame = PlayerInput::new(3, TestInput { inp: 0 });
         queue.add_input(input_wrong_frame); // not fine
     }
 
     #[test]
     #[should_panic]
     fn test_add_input_twice() {
-        let input_size = std::mem::size_of::<u32>();
-        let mut queue = InputQueue::new(input_size);
-        let input = GameInput::new(0, input_size, vec![0]);
-        queue.add_input(input.clone()); // fine
+        let mut queue = InputQueue::<TestConfig>::new();
+        let input = PlayerInput::new(0, TestInput { inp: 0 });
+        queue.add_input(input); // fine
         queue.add_input(input); // not fine
     }
 
     #[test]
     fn test_add_input_sequentially() {
-        let input_size = std::mem::size_of::<u32>();
-        let mut queue = InputQueue::new(input_size);
+        let mut queue = InputQueue::<TestConfig>::new();
         for i in 0..10 {
-            let input = GameInput::new(i, input_size, vec![0; input_size]);
+            let input = PlayerInput::new(i, TestInput { inp: 0 });
             queue.add_input(input);
             assert_eq!(queue.last_added_frame, i);
             assert_eq!(queue.length, (i + 1) as usize);
@@ -282,35 +299,30 @@ mod input_queue_tests {
 
     #[test]
     fn test_input_sequentially() {
-        let input_size = std::mem::size_of::<u32>();
-        let mut queue = InputQueue::new(input_size);
+        let mut queue = InputQueue::<TestConfig>::new();
         for i in 0..10 {
-            let fake_inputs: u32 = i as u32;
-            let serialized_inputs = bincode::serialize(&fake_inputs).unwrap();
-            let input = GameInput::new(i, input_size, serialized_inputs);
-            queue.add_input(input.clone());
+            let input = PlayerInput::new(i, TestInput { inp: i as u8 });
+            queue.add_input(input);
             assert_eq!(queue.last_added_frame, i);
             assert_eq!(queue.length, (i + 1) as usize);
-            let input_in_queue = queue.input(i);
-            assert!(input_in_queue.equal(&input, false));
+            let (input_in_queue, _status) = queue.input(i);
+            assert_eq!(input_in_queue.inp, i as u8);
         }
     }
 
     #[test]
     fn test_delayed_inputs() {
-        let input_size = std::mem::size_of::<u32>();
-        let mut queue = InputQueue::new(input_size);
+        let mut queue = InputQueue::<TestConfig>::new();
         let delay: i32 = 2;
-        queue.set_frame_delay(delay as u32);
+        queue.set_frame_delay(delay as usize);
         for i in 0..10 {
-            let fake_inputs: u32 = i as u32;
-            let serialized_inputs = bincode::serialize(&fake_inputs).unwrap();
-            let input = GameInput::new(i, input_size, serialized_inputs);
-            queue.add_input(input.clone());
+            let input = PlayerInput::new(i, TestInput { inp: i as u8 });
+            queue.add_input(input);
             assert_eq!(queue.last_added_frame, i + delay);
             assert_eq!(queue.length, (i + delay + 1) as usize);
-            let input_in_queue = queue.input(i + delay);
-            assert!(input_in_queue.equal(&input, true));
+            let (input_in_queue, _status) = queue.input(i);
+            let correct_input = std::cmp::max(0, i - delay) as u8;
+            assert_eq!(input_in_queue.inp, correct_input);
         }
     }
 }
