@@ -1,96 +1,82 @@
 use std::collections::HashMap;
 
 use crate::error::GGRSError;
-use crate::frame_info::GameInput;
-use crate::network::udp_msg::ConnectionStatus;
+use crate::frame_info::PlayerInput;
+use crate::network::messages::ConnectionStatus;
 use crate::sync_layer::SyncLayer;
-use crate::{Frame, GGRSRequest, PlayerHandle};
+use crate::{Config, Frame, GGRSRequest, PlayerHandle};
 
 /// During a `SyncTestSession`, GGRS will simulate a rollback every frame and resimulate the last n states, where n is the given check distance.
 /// The resimulated checksums will be compared with the original checksums and report if there was a mismatch.
-pub struct SyncTestSession<T: Clone = Vec<u8>> {
-    num_players: u32,
-    input_size: usize,
+pub struct SyncTestSession<T>
+where
+    T: Config,
+{
+    num_players: usize,
     max_prediction: usize,
     check_distance: usize,
     sync_layer: SyncLayer<T>,
     dummy_connect_status: Vec<ConnectionStatus>,
-    checksum_history: HashMap<Frame, u64>,
+    checksum_history: HashMap<Frame, Option<u128>>,
+    local_inputs: HashMap<PlayerHandle, PlayerInput<T::Input>>,
 }
 
-impl<T: Clone> SyncTestSession<T> {
-    /// Creates a new `SyncTestSession`. During a sync test, GGRS will simulate a rollback every frame and resimulate the last n states, where n is the given `check_distance`.
-    /// During a `SyncTestSession`, GGRS will simulate a rollback every frame and resimulate the last n states, where n is the given check distance.
-    /// The resimulated checksums will be compared with the original checksums and report if there was a mismatch.
-    /// Due to the decentralized nature of saving and loading gamestates, checksum comparisons can only be made if `check_distance` is 2 or higher.
-    /// This is a great way to test if your system runs deterministically. After creating the session, add a local player, set input delay for them and then start the session.
-    /// # Example
-    ///
-    /// ```
-    /// # use ggrs::{GGRSError, SyncTestSession};
-    /// # fn main() -> Result<(), GGRSError> {
-    /// let check_distance : usize = 7;
-    /// let max_pred : usize = 8;
-    /// let num_players : u32 = 2;
-    /// let input_size : usize = std::mem::size_of::<u32>();
-    /// let mut session = SyncTestSession::<Vec<u8>>::new(num_players, input_size, max_pred, check_distance)?;
-    /// # Ok(())
-    /// # }
-    /// ```
-    ///
-    /// # Errors
-    /// - Will return a `InvalidRequestError` if the `check_distance is` higher than or equal to `MAX_PREDICTION_FRAMES`.
-    pub fn new(
-        num_players: u32,
-        input_size: usize,
-        max_prediction: usize,
-        check_distance: usize,
-    ) -> Result<Self, GGRSError> {
-        if check_distance >= max_prediction {
-            return Err(GGRSError::InvalidRequest {
-                info: "Check distance too big.".to_owned(),
-            });
-        }
-        Ok(Self::new_impl(
-            num_players,
-            input_size,
-            max_prediction,
-            check_distance,
-        ))
-    }
-
-    /// Creates a new `SyncTestSession` instance with given values.
-    fn new_impl(
-        num_players: u32,
-        input_size: usize,
+impl<T: Config> SyncTestSession<T> {
+    pub(crate) fn new(
+        num_players: usize,
         max_prediction: usize,
         check_distance: usize,
+        input_delay: usize,
     ) -> Self {
         let mut dummy_connect_status = Vec::new();
         for _ in 0..num_players {
             dummy_connect_status.push(ConnectionStatus::default());
         }
+
+        let mut sync_layer = SyncLayer::new(num_players, max_prediction);
+        for i in 0..num_players {
+            sync_layer.set_frame_delay(i, input_delay);
+        }
+
         Self {
             num_players,
-            input_size,
             max_prediction,
             check_distance,
-            sync_layer: SyncLayer::new(num_players, input_size, max_prediction),
+            sync_layer,
             dummy_connect_status,
-            checksum_history: HashMap::default(),
+            checksum_history: HashMap::new(),
+            local_inputs: HashMap::new(),
         }
     }
 
+    /// Registers local input for a player for the current frame. This should be successfully called for every local player before calling `advance_frame()`.
+    /// If this is called multiple times for the same player before advancing the frame, older given inputs will be overwritten.
+    /// In a sync test, all players are considered to be local, so you need to add input for all of them.
+    ///
+    /// # Errors
+    /// - Returns `InvalidRequest` when the given handle is not valid (i.e. not between 0 and num_players).
+    pub fn add_local_input(
+        &mut self,
+        player_handle: PlayerHandle,
+        input: T::Input,
+    ) -> Result<(), GGRSError> {
+        if player_handle >= self.num_players {
+            return Err(GGRSError::InvalidRequest {
+                info: "The player handle you provided is not valid.".to_owned(),
+            });
+        }
+        let player_input = PlayerInput::<T::Input>::new(self.sync_layer.current_frame(), input);
+        self.local_inputs.insert(player_handle, player_input);
+        Ok(())
+    }
+
     /// In a sync test, this will advance the state by a single frame and afterwards rollback `check_distance` amount of frames,
     /// resimulate and compare checksums with the original states. Returns an order-sensitive `Vec<GGRSRequest>`.
     /// You should fulfill all requests in the exact order they are provided. Failure to do so will cause panics later.
     ///
     /// # Errors
     /// - Returns `MismatchedChecksumError` if checksums don't match after resimulation.
-    pub fn advance_frame(
-        &mut self,
-        all_inputs: &[Vec<u8>],
-    ) -> Result<Vec<GGRSRequest<T>>, GGRSError> {
+    pub fn advance_frame(&mut self) -> Result<Vec<GGRSRequest<T>>, GGRSError> {
         let mut requests = Vec::new();
 
         // if we advanced far enough into the game do comparisons and rollbacks
@@ -110,19 +96,19 @@ impl<T: Clone> SyncTestSession<T> {
             self.adjust_gamestate(frame_to, &mut requests);
         }
 
-        // pass all inputs into the sync layer
-        assert_eq!(self.num_players as usize, all_inputs.len());
-        for (i, input_bytes) in all_inputs.iter().enumerate() {
-            //create an input struct for current frame
-            let input: GameInput = GameInput::new(
-                self.sync_layer.current_frame(),
-                self.input_size,
-                input_bytes.to_vec(),
-            );
-
-            // send the input into the sync layer
-            self.sync_layer.add_local_input(i, input)?;
+        // we require inputs for all players
+        if self.num_players as usize != self.local_inputs.len() {
+            return Err(GGRSError::InvalidRequest {
+                info: "Missing local input while calling advance_frame().".to_owned(),
+            });
         }
+        // pass all inputs into the sync layer
+        for (&handle, &input) in self.local_inputs.iter() {
+            // send the input into the sync layer
+            self.sync_layer.add_local_input(handle, input)?;
+        }
+        // clear local inputs after using them
+        self.local_inputs.clear();
 
         // save the current frame in the syncronization layer
         // we can skip all the saving if the check_distance is 0
@@ -134,9 +120,6 @@ impl<T: Clone> SyncTestSession<T> {
         let inputs = self
             .sync_layer
             .synchronized_inputs(&self.dummy_connect_status);
-        for input in &inputs {
-            assert_eq!(input.frame, self.sync_layer.current_frame());
-        }
 
         // advance the frame
         requests.push(GGRSRequest::AdvanceFrame { inputs });
@@ -156,33 +139,11 @@ impl<T: Clone> SyncTestSession<T> {
         Ok(requests)
     }
 
-    /// Change the amount of frames GGRS will delay the inputs for a player.
-    /// # Errors
-    /// Returns `InvalidHandle` if the provided player handle is higher than the number of players.
-    /// Returns `InvalidRequest` if the provided player handle refers to a remote player.
-    pub fn set_frame_delay(
-        &mut self,
-        frame_delay: u32,
-        player_handle: PlayerHandle,
-    ) -> Result<(), GGRSError> {
-        // player handle is invalid
-        if player_handle > self.num_players as PlayerHandle {
-            return Err(GGRSError::InvalidHandle);
-        }
-        self.sync_layer.set_frame_delay(player_handle, frame_delay);
-        Ok(())
-    }
-
     /// Returns the number of players this session was constructed with.
-    pub fn num_players(&self) -> u32 {
+    pub fn num_players(&self) -> usize {
         self.num_players
     }
 
-    /// Returns the input size this session was constructed with.
-    pub fn input_size(&self) -> usize {
-        self.input_size
-    }
-
     /// Returns the maximum prediction window of a session.
     pub fn max_prediction(&self) -> usize {
         self.max_prediction
@@ -196,18 +157,14 @@ impl<T: Clone> SyncTestSession<T> {
             .retain(|&k, _| k >= oldest_allowed_frame);
 
         match self.sync_layer.saved_state_by_frame(frame_to_check) {
-            Some(latest_cell) => {
-                let latest_state = latest_cell.load();
-
-                match self.checksum_history.get(&latest_state.frame) {
-                    Some(cs) => *cs == latest_state.checksum,
-                    None => {
-                        self.checksum_history
-                            .insert(latest_state.frame, latest_state.checksum);
-                        true
-                    }
+            Some(latest_cell) => match self.checksum_history.get(&latest_cell.frame()) {
+                Some(&cs) => cs == latest_cell.checksum(),
+                None => {
+                    self.checksum_history
+                        .insert(latest_cell.frame(), latest_cell.checksum());
+                    true
                 }
-            }
+            },
             None => true,
         }
     }
