@@ -1,12 +1,9 @@
-use bincode;
 use rand::{prelude::ThreadRng, thread_rng, Rng};
 use std::collections::hash_map::DefaultHasher;
 use std::hash::{Hash, Hasher};
+use std::net::SocketAddr;
 
-use ggrs::{Frame, GGRSRequest, GameInput, GameState, GameStateCell};
-
-pub const INPUT_SIZE: usize = std::mem::size_of::<u32>();
-pub const MAX_PRED_FRAMES: usize = 8;
+use ggrs::{Config, Frame, GGRSRequest, GameStateCell, InputStatus};
 
 fn calculate_hash<T: Hash>(t: &T) -> u64 {
     let mut s = DefaultHasher::new();
@@ -15,19 +12,34 @@ fn calculate_hash<T: Hash>(t: &T) -> u64 {
 }
 
 pub struct GameStub {
-    pub gs: GameStateStub,
+    pub gs: StateStub,
+}
+use bytemuck::{Pod, Zeroable};
+
+#[repr(C)]
+#[derive(Copy, Clone, PartialEq, Pod, Zeroable)]
+pub struct StubInput {
+    pub inp: u32,
+}
+
+pub struct StubConfig;
+
+impl Config for StubConfig {
+    type Input = StubInput;
+    type State = StateStub;
+    type Address = SocketAddr;
 }
 
 impl GameStub {
     #[allow(dead_code)]
     pub fn new() -> GameStub {
         GameStub {
-            gs: GameStateStub { frame: 0, state: 0 },
+            gs: StateStub { frame: 0, state: 0 },
         }
     }
 
     #[allow(dead_code)]
-    pub fn handle_requests(&mut self, requests: Vec<GGRSRequest<GameStateStub>>) {
+    pub fn handle_requests(&mut self, requests: Vec<GGRSRequest<StubConfig>>) {
         for request in requests {
             match request {
                 GGRSRequest::LoadGameState { cell, .. } => self.load_game_state(cell),
@@ -37,26 +49,23 @@ impl GameStub {
         }
     }
 
-    fn save_game_state(&mut self, cell: GameStateCell<GameStateStub>, frame: Frame) {
+    fn save_game_state(&mut self, cell: GameStateCell<StateStub>, frame: Frame) {
         assert_eq!(self.gs.frame, frame);
         let checksum = calculate_hash(&self.gs);
-        let game_state: GameState<GameStateStub> =
-            GameState::<GameStateStub>::new_with_checksum(frame, Some(self.gs), checksum);
-        cell.save(game_state);
+        cell.save(frame, Some(self.gs), Some(checksum as u128));
     }
 
-    fn load_game_state(&mut self, cell: GameStateCell<GameStateStub>) {
-        let game_state: GameState<GameStateStub> = cell.load();
-        self.gs = game_state.data.unwrap();
+    fn load_game_state(&mut self, cell: GameStateCell<StateStub>) {
+        self.gs = cell.load().unwrap();
     }
 
-    fn advance_frame(&mut self, inputs: Vec<GameInput>) {
+    fn advance_frame(&mut self, inputs: Vec<(StubInput, InputStatus)>) {
         self.gs.advance_frame(inputs);
     }
 }
 
 pub struct RandomChecksumGameStub {
-    pub gs: GameStateStub,
+    pub gs: StateStub,
     rng: ThreadRng,
 }
 
@@ -64,13 +73,13 @@ impl RandomChecksumGameStub {
     #[allow(dead_code)]
     pub fn new() -> RandomChecksumGameStub {
         RandomChecksumGameStub {
-            gs: GameStateStub { frame: 0, state: 0 },
+            gs: StateStub { frame: 0, state: 0 },
             rng: thread_rng(),
         }
     }
 
     #[allow(dead_code)]
-    pub fn handle_requests(&mut self, requests: Vec<GGRSRequest<GameStateStub>>) {
+    pub fn handle_requests(&mut self, requests: Vec<GGRSRequest<StubConfig>>) {
         for request in requests {
             match request {
                 GGRSRequest::LoadGameState { cell, .. } => self.load_game_state(cell),
@@ -80,36 +89,32 @@ impl RandomChecksumGameStub {
         }
     }
 
-    fn save_game_state(&mut self, cell: GameStateCell<GameStateStub>, frame: Frame) {
+    fn save_game_state(&mut self, cell: GameStateCell<StateStub>, frame: Frame) {
         assert_eq!(self.gs.frame, frame);
 
-        let random_checksum: u64 = self.rng.gen();
-        cell.save(GameState::new_with_checksum(
-            frame,
-            Some(self.gs),
-            random_checksum,
-        ));
+        let random_checksum: u128 = self.rng.gen();
+        cell.save(frame, Some(self.gs), Some(random_checksum));
     }
 
-    fn load_game_state(&mut self, cell: GameStateCell<GameStateStub>) {
-        self.gs = cell.load().clone().data.expect("No data found.");
+    fn load_game_state(&mut self, cell: GameStateCell<StateStub>) {
+        self.gs = cell.load().expect("No data found.");
     }
 
-    fn advance_frame(&mut self, inputs: Vec<GameInput>) {
+    fn advance_frame(&mut self, inputs: Vec<(StubInput, InputStatus)>) {
         self.gs.advance_frame(inputs);
     }
 }
 
 #[derive(Default, Copy, Clone, Hash)]
-pub struct GameStateStub {
+pub struct StateStub {
     pub frame: i32,
     pub state: i32,
 }
 
-impl GameStateStub {
-    fn advance_frame(&mut self, inputs: Vec<GameInput>) {
-        let p0_inputs: u32 = bincode::deserialize(&inputs[0].buffer).unwrap();
-        let p1_inputs: u32 = bincode::deserialize(&inputs[0].buffer).unwrap();
+impl StateStub {
+    fn advance_frame(&mut self, inputs: Vec<(StubInput, InputStatus)>) {
+        let p0_inputs = inputs[0].0.inp;
+        let p1_inputs = inputs[1].0.inp;
 
         if (p0_inputs + p1_inputs) % 2 == 0 {
             self.state += 2;
